#ifndef NGAJDGHKAJSDFBAJGDBAKJHFBSA
#define NGAJDGHKAJSDFBAJGDBAKJHFBSA

#include <assert.h>
#include "block.h"


// Vector theta is decomposed as
// theta = theta0 + \sum_t theta_t
//
// theta0 corresponds to the special "subproblem0" which is just a decomposition into individual nodes and edges.
//
// Each subproblem (subproblem_t and subproblem_0) has two weights: weight_nodes and weight_edges.
// These weights are used only in AverageXXX() functions (and also RemoveTree() - see below).
// So these weights can be dynamically modified.
//
// When calling AverageXXX(), each node and edge in XXX must be covered by at least one element with positive weight.
//
// It's probably better to store theta in tree subproblems rather than in subproblem0.
// To achieve this, call SetNodeWeights(0) and SetEdgeWeights(0) before calling AverageXXX()
// (but only if each node and edge in XXX is covered by some tree, otherwise the code will complain).


// DT 10/30/2010
#include "DisjointSets.h"
#include <fstream>
#include <vector>
#include <list>
#include <set>
#include "utils.h"
#include "constants_declaration.h"

using namespace std;

typedef enum { 
    STARS_SCHEDULE,
    COMBS_SCHEDULE,
    DYNAMIC_SCHEDULE
} schedule_t;


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// MAIN workhorse class. Holds all the myriad datastructures. Originally written
// by V. Kolmogorov. Later changes and additions were written by D. Tarlow 
// and D. Batra. 

class Graph
{
public:
    struct Tree;
    typedef int NodeId;
    typedef int EdgeId;
    typedef Tree* TreeId;
    
    //////////////////////////////////////////////////////////////////////////////
    /////////////////////////// pairwise terms ///////////////////////////////////
    //////////////////////////////////////////////////////////////////////////////
    
#ifdef TYPE_POTTS
    typedef double V_PTR;
    
    double GET_COST(int xi, int xj, V_PTR V, int Ki, int Kj)
    {
        return (xi == xj) ? 0 : V;
    }
    
    // sets Dj[xj] = min_xi { Di[xi] + weight*V(xi, xj) } - const
    // Returns the const
    double DISTANCE_TRANSFORM(double* Di, double* Dj, V_PTR V, int Ki, int Kj, double weight, int dir, bool if_count=true)
    {
        // DB 01/25 -- keep track of num dist trans
        if (if_count)
            ++num_dist_trans;
        
        int k;
        
        double min = Di[0];
        for (k=1; k<Ki; k++)
        {
            if (min > Di[k]) min = Di[k];
        }
        V *= weight;
        for (k=0; k<Ki; k++)
        {
            Dj[k] = Di[k] - min;
            if (Dj[k] > V) Dj[k] = V;
        }
        
        return min;
    }
    
    // sets xi_opt = arg min_xi { Di[xi] + weight*V(xi, xj) }
    double COMPUTE_MIN(int& xi_opt, double* Di, int xj, V_PTR V, int Ki, int Kj, double weight, int dir)
    {
        int k;
        
        V *= weight;
        
        xi_opt = xj;
        double min = Di[xj] - V;
        
        for (k=0; k<Ki; k++)
        {
            if (min > Di[k]) { min = Di[k]; xi_opt = k; }
        }
        
        return min + V;
    }
    
#else // GENERAL TYPE
    
    typedef double* V_PTR; // V(xi, xj) = V[xi + xj*Ki]
    
    double GET_COST(int xi, int xj, V_PTR V, int Ki, int Kj)
    {
        return V[xi + xj*Ki];
    }
    
    // sets Dj[xj] = min_xi { Di[xi] + weight*V(xi, xj) } - const
    // Returns the const
    double DISTANCE_TRANSFORM(double* Di, double* Dj, V_PTR V, int Ki, int Kj, double weight, int dir, bool if_count=true)
    {
        // DB 01/25 -- keep track of num dist trans
        if (if_count)
            ++num_dist_trans;
        
        int shift0, shift1;
        if (dir == 0) { shift0 = Ki; shift1 = 1;  }
        else          { shift0 = 1;  shift1 = Kj; }
        
        int xi, xj;
        double Dj_min;
        for (xj=0; xj<Kj; xj++)
        {
            double* Vptr = V + xj*shift0;
            double v_min = Di[0] + weight*(*Vptr); Vptr += shift1;
            for (xi=1; xi<Ki; xi++)
            {
                double v = Di[xi] + weight*(*Vptr); Vptr += shift1;
                if (v_min > v) v_min = v;
            }
            Dj[xj] = v_min;
            if (xj == 0 || Dj_min > v_min) Dj_min = v_min;
        }
        for (xj=0; xj<Kj; xj++) Dj[xj] -= Dj_min;
        return Dj_min;
    }
    
    // sets xi_opt = arg min_xi { Di[xi] + weight*V(xi, xj) }
    double COMPUTE_MIN(int& xi_opt, double* Di, int xj, V_PTR V, int Ki, int Kj, double weight, int dir)
    {
        int shift0, shift1;
        if (dir == 0) { shift0 = Ki; shift1 = 1;  }
        else          { shift0 = 1;  shift1 = Kj; }
        
        int xi;
        double* Vptr = V + xj*shift0;
        xi_opt = 0;
        double v_min = Di[0] + weight*(*Vptr); Vptr += shift1;
        for (xi=1; xi<Ki; xi++)
        {
            double v = Di[xi] + weight*(*Vptr); Vptr += shift1;
            if (v_min > v) { xi_opt = xi; v_min = v; }
        }
        return v_min;
    }
#endif
    //////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////////
    
    Graph(int node_num_max, int edge_num_max);
    ~Graph();
    
    // Note: the code will NOT copy arrays D and V, so the user must keep these arrays in the memory.
    // First call to AddNode() returns 0, the second one returns 1, and so on.
    // The same for AddEdge().
    NodeId AddNode(int K, double* D); // K = # labels, D - array of size K.
    EdgeId AddEdge(NodeId i, NodeId j, V_PTR V);
    // must be called before adding any trees. 
    void FinalizeGraph();
    
    EdgeId GetEdge(NodeId i, NodeId j); // returns -1, if there is no edge between i and j
    
    // To add a tree, call AddTree();  ... AddTreeEdge()...AddTreeEdge()... FinalizeTree().
    TreeId AddTree(NodeId root);
    // adds e to tree t. Exactly one endpoint of e must be in t.
    void AddTreeEdge(TreeId t, EdgeId e, bool main_edge=false); // I am not very confident now about this flag, it's better not to use main_edge
    
    // theta_t is set to zero. Use AverageXXX to change that.
    void FinalizeTree(TreeId t); 
    // This function first sets t->weight_nodes = t->weight_edges = 0 and calls AverageTree(t).
    // Therefore, after removing this tree edge node and edge must still be covered by at least one subproblem with positive weight.
    // If, for example, t is the only tree for some edge then you need to call SetEdgeWeights(1) before calling RemoveTree(t).
    void RemoveTree(TreeId t);
    
    // omega \in (0,1] (used only for VNK)
    // Returns lower bound.
    // After this operation t->lb = min Et(x) = \sum_{i\in t} \min theta_i(x_i)   +      \sum_{(i,j)\in t} \min theta_ij(x_i,x_j)
    typedef enum { METHOD_VNK, METHOD_SONTAG } RepTreeMethod;
    double ReparameterizeTree(TreeId t, RepTreeMethod=METHOD_VNK, double omega=1); // returns lower bound
    
    // t=NULL corresponds to subproblem0
    void SetNodeWeights(double w, TreeId t=NULL) { if (t) t->weight_nodes = w; else weight0_nodes = w; }
    void SetEdgeWeights(double w, TreeId t=NULL) { if (t) t->weight_edges = w; else weight0_edges = w; }
    
    void AverageEdge(EdgeId _e, bool compute_lower_bound=false) { AverageEdge(edges + _e, compute_lower_bound); }
    void AverageNode(NodeId _i, bool compute_lower_bound=false) { AverageNode(nodes + _i, compute_lower_bound); }
    void AverageAll(bool compute_lower_bound=false);
    void AverageTree(TreeId t, bool compute_lower_bound=false);
    
    void AverageNonTreeEdges(Tree *t, bool compute_lower_bound=false);  // DT 11/5/2010
    void AverageNonTreeEdgesAndNodes(Tree *t, bool compute_lower_bound=false);  // DT 11/9/2010
    
    void ClearLowerBound() { lb = 0; }
    double GetLowerBound() { return lb; }
    void ClearLowerBound0() { lb0 = 0; }
    double GetLowerBound0() { return lb0; }
    void ClearTreeLowerBound(TreeId t) { t->lb = 0; }
    double GetTreeLowerBound(TreeId t) { return t->lb; }
    
    // if labeling is NULL, then returns the cost of internally stored solution
    double ComputeEnergy(int* labeling = NULL); 
    
    void SerializeMessagesAndTheta0s(ofstream &out);
    void LoadMessagesAndTheta0s(ifstream &in);
    void IncrementMaskUnaries(string mask_filename, double *delta);
    void SetNodeLabel(int node_id, int label, int K);
    
    void Print0();
    void PrintTree(TreeId t);
    
    // DT 10/30/2010
    int ComputeNodeLocalAssignment(int node_id, bool first_iteration=false);
    void ComputeCombinedLPDGs();
    void AddDilatedLPDGs();
    void MakeAllDirty();
    void MakeAllClean();
    void MakeAllLBDirty();
    void MakeAllLBClean();
    
    void ComputeEdgeWTAMeasures();
    void ComputeNodeWTAMeasures();
    
    void AssignRandomEdgeLPDGs();
    
    void ChooseTree(schedule_t schedule, int iteration, int *labeling = NULL);
    void ChooseSpanningTree_Kruskal();
    void ChooseSpanningTree_KruskalNoZeros(double thresh=1e-10);
    void ChooseSpanningTree_HorizontalComb();
    void ChooseSpanningTree_VerticalComb();
    void ChooseSpanningTree_Star(int node_at_center);
    
    void UpdateLastUsedBonuses();
    
    void StoreLabelingInHistories(int *labeling, int max_history_size=5, 
                                  bool is_first_iteration=false);
    
    int *SpanningTreeEdgeIndices() { return spanning_tree_edge_ids; }
    
    int seed_for_random_tree_experiment;
    
    vector<list<int> > edges_adjacent_to_node;
    
    bool edges_are_in_topological_order;
    int kruskal_num_edges_used;
    DisjointSets *kruskal_sets;
    
    // DB 01/25 -- keep track of no of calls to distance transform
    int num_dist_trans;
    
    int size_x;
    int size_y;
    
    
    struct Node;
    struct Edge;
    struct TreeNode;
    struct TreeEdge;
    
    // DT
    double ComputeLPDG(Edge *e, int *labeling = NULL);
    double ComputeNodeLPDG(Node *i, int *labeling);
    double ComputeEdgeWTAMeasure(Edge *e);
    double ComputeNodeWTAMeasure(Node *n);
    
    void DilateLPDGs(int amt, bool is_grid = true);
    
    double *cached_edge_bounds;
    double *cached_node_bounds;
    
    double *lpdgs;
    double *node_lpdgs;
    double *tmp_node_lpdgs;
    double *combined_lpdgs;
    double *last_updated_bonus;
    double *num_updates_node_bonus;
    int *spanning_tree_edge_ids;
    double capacity_scaling_threshold;
    
    
    vector<pair<double, int> > vals_and_indices;  // temporary storage for kruskal
    
    struct Node
    {
        Node() { x = 0; K = 0; D = 0; theta0 = 0; first[0] = 0; first[1] = 0;
            first_tree_node = 0; tmp_count = 0; 
            label_history.push_back(0);
            //label_set.insert(0);
            dirty = true;
            lb_dirty = true;
        }
        
        int x; // solution
        int K; // # labels
        double* D; // unary term - pointer to external array
        double* theta0; // reparameterization
        Edge* first[2];
        TreeNode* first_tree_node;
        int tmp_count; // temporary, used in tree reparameterization;
        bool dirty;
        bool lb_dirty;
        list<int> label_history;
        set<int> label_set;
    };
    
    // reparameterization: if te!=NULL then theta_ij(xi, xj)=0,
    // otherwise   theta_ij(xi, xj) = weight0*V(xi, xj) - M_ij[xj] - M_ji[xi]
    // Note, e->weight0 + \sum_{te \in e} te->weight = 1
    struct Edge
    {
        Node* i[2];
        V_PTR V;
        Edge* next[2];
        TreeEdge* first_tree_edge;
        double weight0;
        double* M[2]; // M[0] - message from i[0] to i[1], M[1] - reverse message
        TreeEdge* te; // if not NULL, then te is the main edge for e;
        // in particular, M[0] and M[1] are the messages in te (same pointers)
    };
    struct Tree
    {
        TreeNode* root;
        double weight_nodes, weight_edges, lb;
        Tree* prev;
        Tree* next;
        double* message_array;
        int message_array_size;
    };
    struct TreeNode
    {
        Tree* t;
        Node* i;
        TreeNode* prev; // list of tree nodes
        TreeNode* next; //   containing i
        
        double* theta;
        
        TreeEdge* parent; // NULL for root
        TreeNode* first_child;
        TreeNode* sibling;
    };
    // reparameterization: theta^t_ij(xi, xj) = weight*V(xi, xj) - M^t_ij[xj] - M^t_ji[xi]
    struct TreeEdge
    {
        Edge* e;
        int dir; // e->i[dir] == child->i, e->i[1-dir] == parent->i
        TreeEdge* prev; // list of tree edges
        TreeEdge* next; //   containing e
        
        TreeNode* parent;
        TreeNode* child;
        
        double weight;
        double* M_to_parent;
        double* M_to_child;
    };
    
    Node* nodes;
    Edge* edges;
    int node_num, node_num_max, edge_num, edge_num_max, K_max;
private:
    Tree* first_tree;
    DBlock<Tree> tree_block;
    DBlock<TreeNode> tree_node_block;
    DBlock<TreeEdge> tree_edge_block;
    
    double weight0_nodes, weight0_edges, lb0, lb;
    
    double* Mbuf; // array of size 4*K_max, plus messages
    
    void AverageEdge(Edge* e, bool compute_lower_bound=false);
    void AverageNode(Node* i, bool compute_lower_bound=false);
};


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// MAIN workhorse function. Implements DTBCA Algorithm as Described in 
// Tarlow, Batra, Kohli, Kolmogorov, ICML '11.  
void dtbca(int alg, int nnodes, int nedges, int *cardinalities, 
           int* edgelist, double* &D, double* &V, int sizeX=0, int sizeY=0);

#endif
