//
//  main.cpp
//  dtbca_public_release
//
//  Created by Dhruv Batra on 8/22/11.
//  Copyright 2011 Dhruv Batra. All rights reserved.
//

#include "main.h"
#include "constants_definition.h"

using namespace std;

// Structure used to parse inputs args
static struct option long_options[] =
{
    {"algorithm", required_argument, 0, 'a'},
    {"MAXITER", required_argument, 0, 'n'},
    {"MAXTIME", required_argument, 0, 't'},
    {"INT_GAP_THRES", required_argument, 0, 'g'},
    {"DUAL_IMPROV_THRES", required_argument, 0, 'd'},
    {"NUM_ITERATIONS_OF_COMBS", required_argument, 0, 'c'},
    {"inputfile", required_argument, 0, 'i'},
    {"outputpath", required_argument, 0, 'o'},
    {"default_potts_lambda", required_argument, 0, 'l'},
    {"energy_type",required_argument,0,'f'},
    {0, 0, 0, 0}
};

int main(int argc, char **argv)
{
    
    //////////////////////////////////////////////////////////////////
    // Parse Arguments
    // getopt_long stores the option index here. 
    int option_index = 0; int c;
    string input_file;
    size_t pos;
    bool first_alg = true; 
    while (1)
    {
        c = getopt_long (argc, argv, "a:n:t:g:d:c:i:o:f:", long_options, &option_index);
        // Detect the end of the options.
        if (c == -1)
            break;
        
        switch (c) {
            case 'a':
                if (first_alg)
                { alg_list.clear(); first_alg = false; }
                alg_list.push_back(atoi(optarg));
                break;
            case 'n':
                MAXITER = atoi(optarg);
                break;
            case 't':
                MAXTIME = atof(optarg);
                break;
            case 'g':
                INT_GAP_THRES = atof(optarg);
                break;
            case 'd':
                DUAL_IMPROV_THRES = atof(optarg);
                break;
            case 'c':
                NUM_ITERATIONS_OF_COMBS = atoi(optarg);
                break;
            case 'i':
            {
                input_file = optarg;
                string tmp_name;
                // find last delimiter; split into path and filename
                splitpath(input_file, inputpath, tmp_name);
                // remove extension from filename                
                splitfname(tmp_name, name, ext);
                break;
            }
            case 'o':
                outputpath = optarg;
                break;
            case 'l':
                default_potts_lambda = atof(optarg);
                break;
            case 'f':
                energytype = optarg;
                break;
            case '?':
                // getopt_long already printed an error so do nothing
                break;
            default:
                CHECK(0, "Unknown Input: Could not parse inputs");
                break;
        }
    }
    
    input_file = fullfile2(inputpath, string(""));
    input_file = input_file + name + ext;

    // Print what the algorithm is running with
    cout << "DTBCA:" << " Alglist: " << alg_list << endl
    << " Input File: " << input_file << endl
    << " Output Path: " << outputpath << endl
    << " Energy File Type: " << energytype << endl
    << " MAXITER: " << MAXITER << " MAXTIME: " << MAXTIME 
    << " INT_GAP_THRES: " << INT_GAP_THRES << " DUAL_IMPROV_THRES: " << DUAL_IMPROV_THRES 
#ifdef TYPE_POTTS
    << " Default_Potts_Lambda: " << default_potts_lambda
#endif
    << endl;
    
    // Print any remaining command line arguments that were not understood.
    if (optind < argc)
    {
        printf ("Followin non-option ARGV-elements were ignored: ");
        while (optind < argc)
            printf ("%s ", argv[optind++]);
        putchar ('\n');
    }

    // Make sure energy file type agrees with compiler flags
    for (int i=0; i!=energytype.size(); ++i)
        energytype[i] = tolower(energytype[i]);
#ifdef TYPE_POTTS
    if (energytype.compare("potts") != 0)
#else
    if ( (energytype.compare(0,3,"gen") != 0) && (energytype.compare(0,3,"uai") != 0) )
#endif
        CHECK(0, "Incompatible or Unknown Energy File Type.");
    
    
    // make directory for output if doesn't exist
    // Assumes *NIX
    DIR *dh = opendir((char*)outputpath.c_str());
    if (dh == NULL)
        mkdir(outputpath.c_str(),0777);

    // Create subdirectory for this instance
    // mkdir assumes *NIX
    outputpath = fullfile2(outputpath, name) + "/";
    mkdir(outputpath.c_str(),0777);
    

    //////////////////////////////////////////////////////////////////
    // Run Algorithms
    // -2 : MPLP
    // -1 : TRW-S
    //  0 : star-TBCA
    //  1 : comb-TBCA
    //  2 : LPDG-based TBCA
    //  3 : WTA(10)-based TBCA
    
    for (int pa=0; pa!=alg_list.size(); ++pa)
    {
        int alg = alg_list[pa];
        
        // A bunch of algorithm directory modify energies; 
        // Thus need to reread from file for each alg. 
        //////////////////////////////////////////////////////////////////
        // Load Energies
        PairwiseEnergy *en; 
#ifdef TYPE_POTTS
        en = new TypePotts;
#else
        if (energytype.compare(0,3,"gen") == 0)
            en = new TypeGeneral;
        else
            en = new TypeUAI;
#endif

        cout << "Loading " << energytype << " energy from file...\t";
        cout.flush();
        MyTimer t; t.tic();
        en->LoadEnergyFromFile((char*) input_file.c_str());
        printf("done\n"); t.toc();

        if (en_round)
            en->RoundEnergy();

#ifdef TYPE_POTTS
        // Algorithms allow cardinalities of each node to be different, so fake create a vector of cardinalities
        int *cardinalities = new int[en->nnodes];
        for (int i=0; i!=en->nnodes; ++i)
            cardinalities[i] = dynamic_cast<TypePotts*>(en)->nlabels;
        
        printf( "#MRF-nodes = %d, #labels = %d, #MRF-edges = %d \n\n", 
               en->nnodes, cardinalities[0], en->nedges);	
#else
        printf( "#MRF-nodes = %d, #MRF-edges = %d, avg node cardinality %f\n", 
               en->nnodes, en->nedges, static_cast<double>(en->sum_of_cardinalities) / en->nnodes);	
        int *cardinalities; 
        if (energytype.compare(0,3,"gen") == 0)
           cardinalities = dynamic_cast<TypeGeneral*>(en)->cardinalities; 
        else
           cardinalities = dynamic_cast<TypeUAI*>(en)->cardinalities;  
#endif

                
        // If not grid, cannot allow certain things
        if (en->nrows*en->ncols<=0) // meaning not a grid
        {
            NUM_ITERATIONS_OF_COMBS = 0; //cannot run combs at the beginning
            CHECK(alg!=1, "Combs schedule is only defined for grid-graph (nrows,ncols cannot be undefined)");
        }
        
        switch (alg) 
        {
            case -2: // MPLP
                printf("Running MPLP:\n");
                //testMPLPGridPotts(en.ncols, en.nrows, en.nedges, cardinalities, en.D, en.V);
                MPLPwrapper(en->nnodes, en->nedges, cardinalities, en->edgelist, en->D, en->V);
                break;
            case -1: // TRW-S
                printf("Running TRW-S:\n");
                TRWSwrapper(en->nnodes, en->nedges, cardinalities, en->edgelist, en->D, en->V);
                break;
            case 0: // Stars
            case 1: // combs
            case 2: // LPDG
            case 3: // WTA           
                printf("Running TBCA alg:%d:\n",alg);
                dtbca(alg, en->nnodes, en->nedges, cardinalities, en->edgelist, en->D, en->V, en->ncols, en->nrows);
                break;
            default:
                EXPECT(0, "Not one of predefined options. Not running any alg.");
                break;
        }
        
        // Garbage collection
        // en is a local variable; has a destructor
#ifdef TYPE_POTTS
        delete [] cardinalities;
        cardinalities = NULL;
#endif
        delete en;
    }
    
}

