//
//  type_general.cpp
//  dtbca_public_release
//
//  Created by Dhruv Batra on 8/25/11.
//  Copyright 2011 Dhruv Batra. All rights reserved.
//
//  Inspired by Vladimir Kolmogorov's typeGeneral.h in TRW-S

#include "type_general.h"

/////////////////////////////////////////////////////////
void TypeGeneral::LoadEnergyFromFile(char *fname)
{
    // Open file
    string input_file(fname);
    ifstream in(input_file.c_str());
    
    CHECK(in.good(),StringPrintf("Could not open file: %s",fname));
    
    // Read size of graph
    in >> nnodes >> nedges >> sum_of_cardinalities >> sum_of_edge_potential_sizes;
    
    // Allocate space
    cardinalities = new Labeltype[nnodes];
    edgelist  = new int [nedges*2];
    D = new double[sum_of_cardinalities];
    V = new double[sum_of_edge_potential_sizes];
    
    // Read node cardinalities
    Labeltype cardinality_i;
    for (int i = 0; i < nnodes; i++) 
    {
        in >> cardinality_i;
        cardinalities[i] = cardinality_i;
        //cout << cardinalities[i] << " ";
    }
    
    // Read node potentials
    int offset = 0;
    double potential_value;
    for (int i = 0; i < nnodes; i++) 
    {
        //cout << i << " (" << cardinalities[i] << ")" << endl;
        for (int k = 0; k < cardinalities[i]; k++) 
        {
            in >> potential_value;
            D[offset++] = potential_value;
            //cout << D[offset-1] << " ";
            
            // keep track of max entry
            if (potential_value > DV_maxval)
                DV_maxval = potential_value;
            // keep track of min entry
            if (potential_value < DV_minval)
                DV_minval = potential_value;

        }
        //cout << endl;
        CHECK(in.good(), "Could not finish reading node potentials. Are you sure this is a typeGeneral energy file?");
    }
    CHECK_EQ(offset, sum_of_cardinalities, "Didn't load right number of node potential entries. Are you sure this is a typeGeneral energy file?");
    
    // Read edge potentials and edgelist
    int n0, n1;
    Labeltype card0, card1;
    offset = 0;
    for (int e = 0; e < nedges; e++) 
    {
        in >> n0 >> n1;
        
        edgelist[2*e] = n0;
        edgelist[2*e+1] = n1;
        
        card0 = cardinalities[n0];
        card1 = cardinalities[n1];
        
        //cout << n0 << " (" << card0 << ") --" << n1 << " (" << card1 << ")" << endl;
        for (int ij = 0; ij < card0 * card1; ij++)
        {
            in >> potential_value;
            V[offset++] = potential_value;
            //cout << V[offset - 1] << " ";

            // keep track of max entry
            if (potential_value > DV_maxval)
                DV_maxval = potential_value;
            // keep track of min entry
            if (potential_value < DV_minval)
                DV_minval = potential_value;
        }
        //cout << endl;
        
        // We allow all edges to share a common edge energy,
        // so if file ends after 1 edge energy, we just replicate parameters.
        // And assume that the graph is a grid (we have to get the edgelist from somewhere)
        if (!in.good() && (e==0)) 
        {
            printf("Parameter sharing detected: all edge potentials will be the same\n");
            
            // Just make sure cardinalities are all the same
            Labeltype card = cardinalities[0];
            for (int i=1; i!=nnodes; ++i) 
                CHECK(card==cardinalities[i], "Parameter sharing implies all cardinalities must be the same");
            
            nrows = n0; ncols = n1;
            CHECK(nrows*ncols==nnodes, "Parameter sharing (for now) implies that the graph must be grid");
            
            // now replicate edge parameters
            for (int e2=1; e2!=nedges; ++e2) 
                memcpy((void*) (V+e2*card*card), (void*) V, card*card*sizeof(double));
            
            // and construct edgelist
            //Construct adjacency list for a 4-connected grid graph
            int ecount = 0, k;
            // Row-Major
            //horizontal edges
            for (int i=0; i!=nrows; ++i) 
                for (int j=0; j!=(ncols-1); ++j) 
                {
                    k = i*ncols + j;
                    edgelist[ecount++] = k;
                    edgelist[ecount++] = k+1;
                }
            //vertical edges
            for (int i=0; i!=(nrows-1); ++i) 
                for (int j=0; j!=ncols; ++j) 
                {
                    k = i*ncols + j;
                    edgelist[ecount++] = k;
                    edgelist[ecount++] = k+ncols;
                }

            offset = sum_of_edge_potential_sizes;
            // now stop reading
            break;
        }
        
        CHECK(in.good(), "Could not finish reading edge potentials. Are you sure this is a typeGeneral energy file?");
    }
    CHECK_EQ(offset, sum_of_edge_potential_sizes, "Didn't load right number of edge potential entries. Are you sure this is a typeGeneral energy file?");
    
    // close file
    in.close();
}

/////////////////////////////////////////////////////////
void TypeGeneral::RoundEnergy()
{
    // Scale to Inttype
    
    // Node energies
    int offset = 0;
    for (int i=0; i!=nnodes; ++i) 
        for (int l=0; l!=cardinalities[i]; ++l) 
        {
            D[offset] = (Inttype) ((D[offset] - DV_minval) * MAX_INT/DV_maxval);
            ++offset;
        }
    // Edge energies
    for (int e=0; e != nedges; ++e) 
    {
        V[e] = (Inttype) ((V[e] - DV_minval) * MAX_INT/DV_maxval);
    }
}
